#!/usr/bin/python
from __future__ import with_statement

BASEDIR = "/home/nasroot/foscripts"
PROGRAM = "delete_all_fs"

import sys
import os

sys.path.append(os.path.join(BASEDIR, "bin/folib"))
import Utility
from Utility import print_status, print_status_warning, print_status_error
import HDIClient
import HDIOperation

sys.path.append(os.path.join(BASEDIR, "bin"))
import Config
import Exceptions
from CommonLogger import log_debug, log_info, log_warning, log_error, log_exception, setup_rotation_logger
import getinfo_common
import backup_sync_common
import setup_common

def delete_fsinfo(fsname, config_dir):
    """ Delete synced information of a file system.

    This function finds synced information of the specified file system
    and delete it.
    This function doesn't care which node the file system came from,
    while backup_sync_common.delete_fsinfo() needs IP address of the
    node that the file system came from..

    Args:
        fsname (str): Name of file system whose synced information will
            be deleted.
        config_dir (str): Directory path that HDI configuration archive
            is stored.
    Returns:
        None
    """
    # Get all synced information
    fsdict_synced = backup_sync_common.get_synced_fslist(config_dir)

    # Find IP address of the node that the file system came from
    for ipaddr in fsdict_synced:
        if fsname in fsdict_synced[ipaddr]:
            ipaddr_of_the_filesystem = ipaddr
            break
    else:
        log_info("Synced information of %s does not exist." % fsname)
        return

    # Delete synced information
    backup_sync_common.delete_fsinfo(fsname, ipaddr_of_the_filesystem, config_dir)

def delete_filesystem_with_resources(fsinfo, config, delete_WORM=False):
    """ Delete a file system with deleting shares and migration tasks on it.

    This function deletes a specified file system with deleting shares and
    migration tasks on the file system.
    A WORM file system will be deleted as well, with a special method.

    Args:
        fsinfo (HDIConfig.FileSystemInfo): Information of the file system
            to be deleted.
        config (Config.Config): Configuration of the scripts.
        delete_WORM (bool): Specify True if you are OK to delete a WORM file
            system.
    Returns:
        None
    """
    print_status("----- Deletion of file system %s -----" % fsinfo.fileSystemName)

    # Setup connection to the node. No SSH credential is required because this is
    # connection to localhost.
    cred = HDIClient.Credential(None, config.rest_user, None, config.rest_passMD5)
    conn = HDIOperation.Connection('localhost', cred, config.retry_count, config.retry_interval, config.timeout)

    # Does this file system exist?
    fsop = HDIOperation.FileSystem(conn)
    try:
        if not fsop.exists(fsinfo.fileSystemName):
            print_status("File system %s does not exist." % fsinfo.fileSystemName)
            # Delete synced information of the file system to clean up the information
            delete_fsinfo(fsinfo.fileSystemName, config.hdi_conf_dir)
            return
    except HDIOperation.OperationError, e:
        print_status_error("Failed to check existence of file system %s. error=%s" % (fsinfo.fileSystemName, str(e)))
        raise Exceptions.OperationError(e.reason, e.code)

    # Delete CIFS shares
    print_status("--- CIFS shares ---")
    cifsop = HDIOperation.CIFSShare(conn)
    try:
        cifslist = cifsop.get_all(fsname=fsinfo.fileSystemName)
    except HDIOperation.OperationError, e:
        print_status_error("Failed to get a list of CIFS shares on %s. error=%s" % (fsinfo.fileSystemName, str(e)))
        raise Exceptions.OperationError(e.reason, e.code)

    for cifsinfo in cifslist:
        print_status("Deleting %s ..." % cifsinfo.cifsShareName)
        try:
            cifsop.delete(cifsinfo)
        except HDIOperation.OperationError, e:
            print_status_error("Failed to delete CIFS share %s. error=%s" % (cifsinfo.cifsShareName, str(e)))
            raise Exceptions.OperationError(e.reason, e.code)
        print_status("%s deleted." % cifsinfo.cifsShareName)

    # Delete NFS shares
    print_status("--- NFS shares ---")
    nfsop = HDIOperation.NFSShare(conn)
    try:
        nfslist = nfsop.get_all(fsname=fsinfo.fileSystemName)
    except HDIOperation.OperationError, e:
        print_status_error("Failed to get a list of NFS shares on %s. error=%s" % (fsinfo.fileSystemName, str(e)))
        raise Exceptions.OperationError(e.reason, e.code)

    for nfsinfo in nfslist:
        print_status("Deleting %s ..." % nfsinfo.exportPoint)
        try:
            nfsop.delete(nfsinfo)
        except HDIOperation.OperationError, e:
            print_status_error("Failed to delete NFS share %s. error=%s" % (nfsinfo.exportPoint, str(e)))
            raise Exceptions.OperationError(e.reason, e.code)
        print_status("%s deleted." % nfsinfo.exportPoint)

    # Delete migration tasks
    print_status("--- Migration tasks ---")
    taskop = HDIOperation.MigrationTask(conn)
    try:
        tasklist = taskop.get_all(fsname=fsinfo.fileSystemName)
    except HDIOperation.OperationError, e:
        print_status_error("Failed to get a list of migration tasks on %s. error=%s" % (fsinfo.fileSystemName, str(e)))
        raise Exceptions.OperationError(e.reason, e.code)

    for taskinfo in tasklist:
        print_status("Deleting %s ..." % taskinfo.task_name)
        try:
            taskop.delete(taskinfo.task_name, fsinfo.fileSystemName)
        except HDIOperation.OperationError, e:
            print_status_error("Failed to delete migration task %s. error=%s" % (taskinfo.task_name, str(e)))
            raise Exceptions.OperationError(e.reason, e.code)
        print_status("%s deleted." % taskinfo.task_name)

    # Delete file system
    print_status("--- File system ---")
    print_status("Deleting %s ..." % fsinfo.fileSystemName)
    fsop = HDIOperation.FileSystem(conn)
    if 'WORM' in fsinfo.usage:
        # WORM deletion is allowed?
        if not delete_WORM:
            print_status("WORM is enabled on %s but not allowed to delete. usage=%s" % (fsinfo.fileSystemName, fsinfo.usage))
            raise Exceptions.NoExecution()

        ### Special method

        # Mount the file system if not mounted
        log_info("Mounting %s if not mounted ..." % fsinfo.fileSystemName)
        try:
            if not fsop.is_mounted(fsinfo.fileSystemName):
                fsop.mount(fsinfo.fileSystemName)
        except HDIOperation.OperationError, e:
            log_error("Failed to mount %s. error=%s" % (fsinfo.fileSystemName, str(e)))
            raise Exceptions.OperationError(e.reason, e.code)

        # Get LV path of the file system
        log_info("Getting LV path for file system %s ..." % fsinfo.fileSystemName)
        cmd = "/bin/df -P"
        log_info("Command-Exec-Start: %s" % cmd)
        ret, out, err = Utility.run(cmd)
        log_info("Command-Exec-End: %s, ret=%s, err=%s" % (cmd, ret, err))
        for line in out.split('\n'):
            els = line.split()
            if els[5] == "/mnt/%s" % fsinfo.fileSystemName:
                lvpath = els[0]
                break
        else:
            # LV of the file system was not found
            log_error("LV path for file system %s was not found. df results: %s" % (fsinfo.fileSystemName, out.split('\n')))
            raise Exceptions.OperationError("LV path for file system %s was not found." % fsinfo.fileSystemName, ret)

        # Unmount the file system
        log_info("Unmounting %s ..." % fsinfo.fileSystemName)
        try:
            fsop.unmount(fsinfo.fileSystemName)
        except HDIOperation.OperationError, e:
            log_error("Failed to unmount %s. error=%s" % (fsinfo.fileSystemName, str(e)))
            raise Exceptions.OperationError(e.reason, e.code)

        # Initialize the LV
        log_info("Initializing LV %s for %s ..." % (lvpath, fsinfo.fileSystemName))
        cmd = "sudo maintexec %s %s %s" % (os.path.join(BASEDIR, "int/int_init_lv.sh"), fsinfo.fileSystemName, lvpath)
        log_info("Command-Exec-Start: %s" % cmd)
        ret, out, err = Utility.run(cmd)
        log_info("Command-Exec-End: %s, ret=%s" % (cmd, ret))
        if ret != 0:
            # Error messages of int_init_lv.sh is output to stdout
            log_error("Failed to Initialize LV. ret=%s, error=%s" % (ret, out))
            raise Exceptions.OperationError("Failed to initialize LV for %s. error=%s" % (fsinfo.fileSystemName, out), ret)

        # Delete the file system
        try:
            fsop.delete(fsinfo.fileSystemName)
        except HDIOperation.OperationError, e:
            print_status_error("Failed to delete %s. error=%s" % (fsinfo.fileSystemName, str(e)))
            raise Exceptions.OperationError(e.reason, e.code)

    else:
        ### Normal method

        # Unmount the file system if mounted
        try:
            if fsop.is_mounted(fsinfo.fileSystemName):
                fsop.unmount(fsinfo.fileSystemName)
        except HDIOperation.OperationError, e:
            print_status_error("Failed to unmount %s. error=%s" % (fsinfo.fileSystemName, str(e)))
            raise Exceptions.OperationError(e.reason, e.code)

        # Delete the file system
        try:
            fsop.delete(fsinfo.fileSystemName)
        except HDIOperation.OperationError, e:
            print_status_error("Failed to delete %s. error=%s" % (fsinfo.fileSystemName, str(e)))
            raise Exceptions.OperationError(e.reason, e.code)

    print_status("%s deleted." % fsinfo.fileSystemName)

    # Delete synced file system information
    delete_fsinfo(fsinfo.fileSystemName, config.hdi_conf_dir)

def main():
    # Enable Ctrl-C interruption
    Utility.enable_interrupt()

    # Load configs
    config = Config.Config()

    # Setup logging
    log_file_path = os.path.join(config.log_dir, config.common_log_file)
    setup_rotation_logger(__name__, config.log_level, log_file_path, config.log_size, config.log_gen)

    log_info("----------------------- %s --------------------------" % PROGRAM)

    # Check resource group status
    print_status("Checking resource group status ...")
    if not getinfo_common.is_running_healthily(config):
        print_status_warning("The node is not running healthily. Stop execution.")
        raise Exceptions.NoExecution()

    # Get file system information synced from other nodes
    fsdict_synced = backup_sync_common.get_synced_fslist(config.hdi_conf_dir)

    # Prepare menu
    OP_ALL = 1
    OP_SPEC = 2
    OP_EXIT = 0
    menu = [
        (OP_ALL, "All file systems"),
        (OP_SPEC, "Specify a file system(s)")
    ]

    idx = 3
    map_idx_to_ipaddr = {}
    for ipaddr in fsdict_synced:
        menu.append((idx, "File systems from %s" % ipaddr))
        map_idx_to_ipaddr[idx] = ipaddr
        idx += 1

    menu.append((OP_EXIT, "Exit"))

    # Select file systems to be deleted

    print
    print ">>> Specify file systems you want to delete."
    selected_list = Utility.select_items("Select file systems.", menu, title="Menu", onlyone=True)
    if selected_list is None:
        log_warning("Selecting a menu item did not succeed.")
        print "Process canceled."
        raise Exceptions.NoExecution()

    selected = selected_list[0]
    if selected == OP_EXIT:
        raise Exceptions.NoExecution()
    elif selected == OP_ALL:
        # All file systems
        # Get file systems on this node
        print
        print "Getting a list of file systems on this node ..."
        fs_list = getinfo_common.get_local_file_systems(config)
        fs_list_to_be_deleted = fs_list
        if len(fs_list) == 0:
            print
            print "No file systems on this node."
            raise Exceptions.NoExecution()
    elif selected == OP_SPEC:
        # Select a file system(s)
        # Get fiel systems on this node
        print
        print "Getting a list of file systems on this node ..."
        fs_list = getinfo_common.get_local_file_systems(config)
        if len(fs_list) == 0:
            print
            print "No file systems on this node."
            raise Exceptions.NoExecution()

        # Build a selection menu
        fs_menu = []
        for i, fsinfo in enumerate(fs_list):
            idx = i + 1
            fs_menu.append((idx, fsinfo.fileSystemName))

        # Select a file system(s) to be deleted
        selected_fsidx_list = Utility.select_items("Select a file system (separate with comma for multiple selection)", fs_menu, title="File systems on this node")
        if selected_fsidx_list is None:
            log_warning("Selecting a file system did not succeed.")
            print "Process canceled."
            raise Exceptions.NoExecution()
        fs_list_to_be_deleted = [fs_list[idx - 1] for idx in selected_fsidx_list]
    else:
        # Get file systems from the selected node
        ipaddr = map_idx_to_ipaddr[selected]
        fsname_list_to_be_deleted = fsdict_synced[ipaddr]
        fs_list_to_be_deleted = [backup_sync_common.get_synced_fsinfo(fsname, ipaddr, config.hdi_conf_dir) for fsname in fsname_list_to_be_deleted]

    # Confirmation for deletion file systems
    print
    print "File systems to be deleted:"
    for fsinfo in fs_list_to_be_deleted:
        print "\t%s" % fsinfo.fileSystemName

    print
    print "----------------------------------"
    print "CAUTION!!"
    print "This operation will delete all file systems above, shares and migration tasks"
    print "on the file systems."
    print "----------------------------------"
    print
    answer = Utility.yes_or_no_full_confirm("Do you really want to delete the file systems?")
    if answer is not True:
        print "Process canceled."
        raise Exceptions.NoExecution()

    # Confirmation for WORM file system deletion
    worm_count = 0
    for fsinfo in fs_list_to_be_deleted:
        if 'WORM' in fsinfo.usage:
            worm_count += 1
    log_info("%s WORM file systems to be deleted." % worm_count)

    if worm_count > 0:
        print
        print "There is %s WORM file system(s) to be deleted." % worm_count
        print
        answer = Utility.yes_or_no_full("Do you really want to delete WORM file systems?")
        if answer is not True:
            print "Process canceled."
            raise Exceptions.NoExecution()
        reason = Utility.my_input("Reason for deletion: ")
        if reason is None or len(reason.strip()) == 0:
            print "You must enter reason to delete WORM file systems."
            print "Process canceled."
            raise Exceptions.NoExecution()

        log_info("Reason for deletion of WORM file systems: %s" % reason)

        confirm = Utility.my_input("Are you sure you want to proceed? (YES/no): ")
        if confirm != "YES":
            print "Process canceled."
            raise Exceptions.NoExecution()

    # Disable backup_configs if this is primary node
    print_status("Checking if I am primary node or not ...")
    myself = setup_common.get_my_ip_address(config)
    if myself is None:
        # Warn just in case though it's not a critical problem for just deleting file systems
        print_status_warning("Couldn't find myself in pair configuration.")
    elif setup_common.is_primary(myself, config.fo_conf_dir):
        print_status("Disabling execution of backup_configs to avoid backup empty configuration ...")
        backup_sync_common.disable_backup(config.fo_conf_dir)

    # Delete resources for each file system
    for fsinfo in fs_list_to_be_deleted:
        try:
            delete_filesystem_with_resources(fsinfo, config, delete_WORM=True)
        except Exception, e:
            print_status_error("Failed to delete %s. Continue to delete other file systems. error=%s" % (fsinfo.fileSystemName, str(e)))

    print_status("Completed.")

if __name__ == '__main__':
    try:
        main()
    except Exceptions.NoExecution:
        sys.exit(0)
    except KeyboardInterrupt:
        print "Interrupted."
        log_warning("Interrupted.")
        sys.exit(2)
    except Config.BaseError, e:
        print "Unable to load script configuration: %s" % str(e)
        sys.exit(10)
    except Exceptions.ConfigurationError, e:
        log_error("Unexpected configuration: %s" % str(e))
        sys.exit(20)
    except Exceptions.ConnectionError, e:
        log_error("A connection error occurred: %s" % str(e))
        sys.exit(70)
    except Exceptions.OperationError, e:
        log_error("An operation error occurred: %s" % str(e))
        sys.exit(90)
    except EnvironmentError, e:
        log_exception("A system error occurred: %s", e)
        sys.exit(99)
    except Exception, e:
        log_exception("An internal error occurred: %s", e)
        sys.exit(99)

    sys.exit(0)
