#!/usr/bin/python
# All Rights Reserved. Copyright (c) 2018, Hitachi, Ltd.       #
# Licensed Material of Hitachi, Ltd.                           #
# Reproduction, use, modification or disclosure otherwise than #
# permitted in the License Agreement is strictly prohibited.   #

from __future__ import with_statement
import os
import errno
import base64
import urllib
import urllib2
import uuid

import Utility

try:
    from CommonLogger import log_debug, log_info, log_warning, log_error
except ImportError:
    # To make this module handy and portable
    def log_debug(msg): pass
    def log_info(msg): pass
    def log_warning(msg): print msg
    def log_error(msg): print msg

class BaseError(Exception):
    def __init__(self, reason, code=None):
        self.reason = reason.strip()
        self.code = code

    def __str__(self):
        if self.code is None:
            return "%s" % self.reason
        else:
            return "%s (code=%s)" % (self.reason, self.code)

class ConnectionError(BaseError):
    pass

class ParameterError(BaseError):
    pass

class HTTPMethod(object):
    # python2.5 doesn't have Enum...
    HEAD = 0
    GET = 1
    PUT = 2
    POST = 3
    DELETE = 4

    @classmethod
    def tostr(cls, method):
        if method == cls.HEAD:
            return "HEAD"
        elif method == cls.GET:
            return "GET"
        elif method == cls.PUT:
            return "PUT"
        elif method == cls.POST:
            return "POST"
        elif method == cls.DELETE:
            return "DELETE"
        else:
            return "UNKNOWN"

class Response(object):
    def __init__(self):
        self.code = None
        self.msg = None
        self.headers = {}   # This header is a dictionary as a receiver concatenates multiple same headers into a single header
        self.body = None

    def info(self):
        return self.headers

    def read(self):
        return self.body

class MyCurl(object):
    def __init__(self):
        self.url = None
        self.params = None
        self.method = HTTPMethod.GET
        self.proxy = None
        self.retry_count = 0
        self.retry_interval = 5
        self.timeout = 0
        self.headers = []   # This headers is a list of tuples as a sender can send the same header multiple times
        self.file_to_send = None
        self.data_to_send = None
        self.to_file = None

    def set_url(self, url):
        self.url = url

    def set_params(self, params):
        self.params = params

    def set_proxy(self, host, port, user=None, rawpass=None):
        self.proxy = {'host': host, 'port': port, 'user': user, 'pass': rawpass}

    def set_method(self, method):
        self.method = method

    def set_retry(self, count, interval):
        self.retry_count = count
        self.retry_interval = interval

    def set_timeout(self, timeout):
        self.timeout = timeout

    def set_data(self, data):
        self.data_to_send = data
        self.file_to_send = None

    def set_file(self, from_file):
        self.file_to_send = from_file
        self.data_to_send = None

    def set_output_file(self, to_file):
        self.to_file = to_file

    def add_header(self, header):
        # header is a tuple
        self.headers.append(header)

    def perform(self):
        # Build curl options
        retry_options = self._build_retry_options()
        timeout_options = self._build_timeout_options()
        proxy_options = self._build_proxy_options()
        proxy_options_hidden = self._build_proxy_options(hidden=True)
        header_options = self._build_header_options()
        method_options = self._build_method_options()
        param_options = self._build_param_options()
        # Record status code and header output to a file
        header_output_file = "/tmp/curl_header_%s" % str(uuid.uuid4())

        # Execute curl
        curl_cmd = "curl -s -S -k --dump-header %s %s %s %s %s %s '%s'" % (header_output_file, retry_options, timeout_options, proxy_options, header_options, method_options, self.url + param_options)
        curl_cmd_hidden = "curl -s -S -k --dump-header %s %s %s %s %s %s '%s'" % (header_output_file, retry_options, timeout_options, proxy_options_hidden, header_options, method_options, self.url + param_options)
        log_debug("Command-Exec-Start: %s" % curl_cmd_hidden)
        ret, out, err = Utility.run(curl_cmd)
        log_debug("Command-Exec-End: %s, ret=%s" % (curl_cmd_hidden, ret))
        if ret != 0:
            log_error("curl finished with an error. ret=%s, err=%s" % (ret, err))
            raise ConnectionError(err, ret)

        # Read status code and headers
        # Multiple status code and headers can be seen in case proxy authentication is executed, "100 Continue", etc.
        # Remembers only the last status code and headers
        version = code = message = None
        headers = {}
        try:
            with open(header_output_file, "r") as f:
                complete = True
                for line in f:
                    log_debug("Header line: %s" % line.strip())
                    if line == '\r\n':
                        complete = True
                        log_debug("HTTP Response (Temporary): code=%s, message=%s, headers=%s" % (code, message, headers))
                    elif complete is True:
                        # The first line (after an empty line)
                        version, code, message = self.parse_status(line)
                        complete = False
                        headers = {}
                    else:
                        hdr = self.parse_header(line)
                        if hdr[0].lower() in [header.lower() for header in headers.keys()]:
                            headers[hdr[0]] += ", %s" % hdr[1]
                        else:
                            headers[hdr[0]] = hdr[1]
        except EnvironmentError, e:
            log_error("Failed to open/read HTTP response header information. file=%s, error=%s" % (header_output_file, str(e)))
            raise

        # Remove the temporary file
        try:
            os.remove(header_output_file)
        except EnvironmentError, e:
            if e.errno == errno.ENOENT:
                log_warning("HTTP response header information file has already gone? Continue... file=%s" % header_output_file)
            else:
                log_error("Failed to delete HTTP response header information file. file=%s. error=%s" % (header_output_file, str(e)))
                raise

        response = Response()
        response.code = code
        response.msg = message
        response.headers = headers
        response.body = out

        log_debug("HTTP Response: code=%s, message=%s, headers=%s" % (response.code, response.msg, response.headers))
        log_debug("HTTP Response body: %s" % response.body)
        return response

    def _build_param_options(self):
        # It seems urllib.urlencode() can't handle a single parameter like "prettyprint"
        if self.params is None or len(self.params) == 0:
            return ""
        params = self.params.copy()
        params_single = []
        for key in params:
            if params[key] is None:
                # Single parameter
                params_single.append(key)
        for key in params_single:
            # Remove single parameters from the dict
            params.pop(key)

        # TODO: Maybe we want to urlencode for params_single as well if it contains non-ascii characters
        params_string = urllib.urlencode(params) + '&' + '&'.join(params_single)
        return '?' + params_string.strip('&')   # Remove the first '&' when params is empty

    def _build_header_options(self):
        if len(self.headers) == 0:
            return ""
        options = ""
        for name, value in self.headers:
            options += "-H \"%s: %s\" " % (name, value)

        return options

    def _build_proxy_options(self, hidden=False):
        if self.proxy is None:
            return ""
        options = "--proxy-anyauth "
        options += "--proxy %s:%s " % (self.proxy['host'], int(self.proxy['port']))
        if self.proxy['user'] is not None:
            if hidden:
                options += "--proxy-user %s:%s" % (self.proxy['user'], "***")
            else:
                options += "--proxy-user %s:%s" % (self.proxy['user'], self.proxy['pass'])

        return options

    def _build_method_options(self):
        if self.method == HTTPMethod.HEAD:
            options = "-I"
        elif self.method == HTTPMethod.GET:
            if self.to_file is not None:
                options = "-o %s" % self.to_file
            else:
                options = ""
        elif self.method == HTTPMethod.PUT:
            if self.file_to_send is not None:
                options = "-T %s" % self.file_to_send
            elif self.data_to_send is not None:
                options = "-d \"%s\" -X PUT" % self.data_to_send
            else:
                raise ParameterError("No data source for PUT request.")
        elif self.method == HTTPMethod.POST:
            if self.file_to_send is not None:
                options = "-d @%s" % self.file_to_send
            elif self.data_to_send is not None:
                options = "-d \"%s\"" % self.data_to_send
            else:
                options = "-X POST"
        elif self.method == HTTPMethod.DELETE:
            options = "-X DELETE"

        return options

    def _build_retry_options(self):
        if self.retry_count == 0:
            options = ""
        else:
            options = "--retry %s --retry-delay %s" % (self.retry_count, self.retry_interval)

        return options

    def _build_timeout_options(self):
        if self.timeout == 0:
            options = ""
        else:
            options = "--max-time %s" % self.timeout

        return options

    @staticmethod
    def parse_status(line):
        try:
            # ex) HTTP/1.1 200 OK
            version, status, message = line.split(None, 2)
        except ValueError:
            try:
                # Message might be empty
                version, status = line.split(None, 1)
                message = ""
            except ValueError:
                # This may not be a status line or it may be HTTP 0.9
                raise ConnectionError("Invalid HTTP status line. line=%s" % line, -1)

        if not version.startswith("HTTP/"):
            raise ConnectionError("Version doesn't start with HTTP/ in HTTP status line. line=%s" % line, -1)

        if not status.isdigit():
            raise ConnectionError("Status code is not a digit in HTTP status line. line=%s" % line, -1)
        status = int(status)
        if status < 100 or status > 999:
            raise ConnectionError("Status code is not a valid HTTP code. line=%s" % line, -1)

        return (version, status, message.strip())

    @staticmethod
    def parse_header(line):
        i = line.find(':')
        if i <= 0:
            raise ConnectionError("Invalid header format in HTTP header. line=%s" % line, -1)
        name = line[:i].strip()
        value = line[i+1:].strip()

        return (name, value)

class HttpClient(object):
    def __init__(self, baseurl):
        self.baseurl = baseurl
        self.headers = []   # This headers is a list of tuples as a sender can send the same header multiple times
        self.params = {}
        self.data = None
        self.to_file = None
        self.retry_count = 0
        self.retry_interval = 0
        self.timeout = 0
        self.proxy = None

    def add_header(self, header):
        # header must be a tuple
        self.headers.append(header)

    def add_param(self, param):
        # param must be a dictionary
        self.params.update(param)

    def add_data(self, data):
        self.data = {'data': data, 'file': None}

    def add_file(self, from_file):
        self.data = {'data': None, 'file': from_file}

    def set_to(self, to_file):
        self.to_file = to_file

    def set_retry(self, count, interval):
        self.retry_count = count
        self.retry_interval = interval

    def set_timeout(self, timeout):
        self.timeout = timeout

    def set_proxy(self, host, port, username=None, password=None):
        self.proxy = {'host': host, 'port': port, 'user': username, 'pass': password}

    def head(self, path):
        response = self._http_request_with_retry(HTTPMethod.HEAD, path)
        ret = {
            'code': response.code,
            'message': response.msg,
            'headers': response.info(),
            'body': None,
        }
        return ret

    def get(self, path, to_file=None):
        if to_file is not None:
            self.set_to(to_file)

        response = self._http_request_with_retry(HTTPMethod.GET, path)
        if to_file is not None:
            body = None
        else:
            body = response.read()
        ret = {
            'code': response.code,
            'message': response.msg,
            'headers': response.info(),
            'body': body,
        }
        return ret

    def put(self, path, from_file=None, data=None):
        if data is None and from_file is None:
            raise ParameterError("Not data source for PUT request.")

        if data is not None:
            self.add_data(data)
        else:
            if not os.path.exists(from_file):
                raise ParameterError("The file to be uploaded by PUT request does not exist. file=%s" % from_file)

            self.add_file(from_file)

        response = self._http_request_with_retry(HTTPMethod.PUT, path)
        ret = {
            'code': response.code,
            'message': response.msg,
            'headers': response.info(),
            'body': response.read()
        }
        return ret

    def post(self, path, from_file=None, data=None, to_file=None):
        if data is not None:
            self.add_data(data)
        elif from_file is not None:
            if not os.path.exists(from_file):
                raise ParameterError("The file to be posted by POST request does not exist. file=%s" % from_file)

            self.add_file(from_file)

        if to_file is not None:
            self.set_to(to_file)

        response = self._http_request_with_retry(HTTPMethod.POST, path)
        if to_file is not None:
            body = None
        else:
            body = response.read()
        ret = {
            'code': response.code,
            'message': response.msg,
            'headers': response.info(),
            'body': body
        }
        return ret

    def delete(self, path):
        response = self._http_request_with_retry(HTTPMethod.DELETE, path)
        ret = {
            'code': response.code,
            'message': response.msg,
            'headers': response.info(),
            'body': response.read()
        }
        return ret

    def _http_request_with_retry(self, method, path):
        # Use curl. urllib2 is no longer used because it doesn't support TLS1.2...

        curl = MyCurl()

        url = self.baseurl + urllib2.quote(path)
        curl.set_url(url)

        if self.proxy is not None:
            if self.proxy['user'] is None:
                curl.set_proxy(self.proxy['host'], self.proxy['port'])
            else:
                curl.set_proxy(self.proxy['host'], self.proxy['port'], self.proxy['user'], base64.b64decode(self.proxy['pass']))

        curl.set_params(self.params)

        curl.set_method(method)
        if method == HTTPMethod.GET:
            if self.to_file is not None:
                curl.set_output_file(self.to_file)
        elif method == HTTPMethod.PUT or method == HTTPMethod.POST:
            if method == HTTPMethod.PUT and self.data is None:
                raise ParameterError("No data source for PUT request.")
            if self.data is not None:
                if self.data['data'] is not None:
                    curl.set_data(self.data['data'])
                else:
                    curl.set_file(self.data['file'])

        for hdr in self.headers:
            curl.add_header(hdr)

        curl.set_retry(self.retry_count, self.retry_interval)
        curl.set_timeout(self.timeout)

        response = curl.perform()

        return response
