#!/usr/bin/python
from __future__ import with_statement

BASEDIR = "/home/nasroot/foscripts"
PROGRAM = "register_filesystem_failover"

import sys
import os
import getopt
import fcntl

sys.path.append(os.path.join(BASEDIR, "bin/folib"))
import Utility
from Utility import print_status
import FailoverConfig
import HDIClient
import HDIOperation

sys.path.append(os.path.join(BASEDIR, "bin"))
import Config
import Exceptions
import Alert
import HDIConfigArchive
from CommonLogger import log_debug, log_info, log_warning, log_error, log_exception, setup_rotation_logger
import failover_common
import lock_common
import backup_sync_common
import setup_common

# Global variable to save information
myself_saved = None
fs_global = None

def usage():
    print "Usage:"
    print "    sudo maintexec %s --source <IP address of source node>" % PROGRAM
    print "                      --file-system <file system name>"
    print "                      [--storage-policy <storage policy string>]"
    print "                      [--failback]"
    print "    sudo maintexec %s -h|--help" % PROGRAM

def get_myself(config):
    global myself_saved
    if myself_saved is None:
        myself_saved = setup_common.get_my_ip_address(config)
    return myself_saved

def register_filesystem_failover(source, fs, policy_string, config_dir, config, is_failback=False):
    """ Register a failover job.

    A failover job is registered in a file, as one job per file.
    Each job file has a different number (serial number of job files).

    This function creates a new job file and record the provided information
    (source node IP address, target fiel system, policy string) to the file.

    A created job file contains explicit file system name and storage policy.
    If fs is "<ALL>", this function downloads the configuration archive of
    the source node and get the actual list of file systems and creates a
    job file for each file system.

    If policy_string is None, this function retrieves a storage policy that
    is applied to the file system.

    This function acquires an exclusive lock to avoid confliction of creating
    a job file with the same number.

    Args:
        source (str): IP address of the failover source node.
        fs (str): File system name to be failed over.
        policy_string (str): String expression of the storage policy to be
            applied to the file system.
            None can be specified if pre-configured policy is used.
        config_dir (str): Failover configuration directory path.
        config (Config.Config): Configuration of the scripts.
        is_failback (bool): True if this is a failover job for failback.
    Returns:
        None
    """
    global fs_global
    if fs == '<ALL>':
        # Need to get list of file systems on the source node, and create
        # a failover job for each file system.
        print_status("Getting information of all file systems on %s ..." % source)

        # Prepare HCP connection
        print_status("Loading HCP configuration ...")
        cred = HDIClient.Credential(None, config.rest_user, None, config.rest_passMD5)
        conn = HDIOperation.Connection('localhost', cred, config.retry_count, config.retry_interval, config.timeout)
        try:
            hcpop = HDIOperation.HCPSetting(conn)
            hcp_info = hcpop.get()
        except HDIOperation.OperationError, e:
            log_error("Failed to get HCP information on the node. error=%s" % str(e))
            raise Exceptions.OperationError("Failed to get HCP configuration on the node. host=%s, error=%s" % (conn.host, e.reason), e.code)

        # Download configuration archive of the source node
        archive_path = backup_sync_common.download_configs(source, config.hdi_conf_dir, hcp_info, config)
        archive = HDIConfigArchive.HDIConfigArchive()
        archive.load(archive_path)

        # Get list of file systems from the archive
        fs_list = archive.fslist
        log_info("File systems on %s: %s" % (source, [fsinfo.fileSystemName for fsinfo in fs_list]))

        # Create a job file for each file system.
        # Storage policy is always None for "<ALL>" case, unless this is failback.
        for fsinfo in fs_list:
            fsname = fsinfo.fileSystemName
            fs_global = fsname
            # Skip registration of a file system if it's not supported
            if not fsinfo.is_supported():
                print_status("Skip registration of %s because it is not supported." % fsname)
                continue
            if is_failback:
                # In failback case, only file systems that were failed over from this node will be restored.
                # Check if the file system was failed over from this node.
                myself = get_myself(config)
                if myself not in archive.synced_fslist.keys() or fsname not in archive.synced_fslist[myself].keys():
                    print_status("Skip registration of %s because it was not failed over from this node." % fsname)
                    continue
                # Specify 'failback' policy if this is failback.
                policy = FailoverConfig.StoragePolicy(fsname, policy=FailoverConfig.StoragePolicy.FAILBACK)
                policy_string = policy.to_string()
            else:
                # Policy is None for <ALL> file system for failover.
                policy_string = None
            print_status("Registering a failover job for %s ..." % fsname)
            create_job_file_with_lock(source, fsname, policy_string, config_dir)
    else:
        # Create a job file for the specified file system.
        fsname = fs
        fs_global = fsname
        # Override the specified policy with 'failback' policy if this is failback.
        if is_failback:
            policy = FailoverConfig.StoragePolicy(fsname, policy=FailoverConfig.StoragePolicy.FAILBACK)
            policy_string = policy.to_string()
        print_status("Registering a failover job for %s ..." % fsname)
        create_job_file_with_lock(source, fsname, policy_string, config_dir)

def create_job_file_with_lock(source, fs, policy_string, config_dir):
    log_info("Creating a job file. source=%s, fs=%s, policy_string=%s, config_dir=%s" % (source, fs, policy_string, config_dir))
    lock_file = lock_common.get_register_lock_file()
    with open(lock_file, 'w') as locker:
        # Acquire lock for failover job entries
        fcntl.flock(locker, fcntl.LOCK_EX)
        # Record the failover job
        failover_common.create_job_file(fs, source, fs, policy_string, config_dir)
        # Release lock
        fcntl.flock(locker, fcntl.LOCK_UN)

def main():
    # Enable Ctrl-C interruption
    Utility.enable_interrupt()

    # Load configs
    config = Config.Config()

    # Setup logging
    log_file_path = os.path.join(config.log_dir, config.common_log_file)
    setup_rotation_logger(__name__, config.log_level, log_file_path, config.log_size, config.log_gen)

    log_info("----------------------- %s --------------------------" % PROGRAM)

    # Parse options
    try:
        short_opts = "h"
        long_opts = ["id=", "source=", "file-system=", "storage-poliy=", "failback", "help"]
        opts, args = getopt.getopt(sys.argv[1:], short_opts, long_opts)
    except getopt.GetoptError, e:
        raise Exceptions.OptionError(e)

    source = None
    fs = None
    policy_string = None
    is_failback = False

    for o, a in opts:
        if o == "--id":
            Utility.ID = a
        if o == "--source":
            source = a
        if o == "--file-system":
            fs = a
            global fs_global
            fs_global = a
        if o == "--storage-policy":
            policy_string = a
        if o == "--failback":
            is_failback = True
        if o in ("-h", "--help"):
            raise Exceptions.HelpOption()

    if source is None or fs is None:
        raise Exceptions.OptionError("Source IP address and file system name are mandatory.")

    if policy_string is not None:
        # Try to create StoragePolicy instance to validate policy_string format
        try:
            notused = FailoverConfig.StoragePolicy(fs, policy_string=policy_string)
        except FailoverConfig.FormatError, e:
            raise Exceptions.OptionError("Invalid storage policy string: %s" % str(e))

    # Register the failover job
    register_filesystem_failover(source, fs, policy_string, config.fo_conf_dir, config, is_failback)

if __name__ == '__main__':
    try:
        main()
    except Exceptions.HelpOption:
        usage()
        sys.exit(0)
    except Exceptions.OptionError, e:
        log_error("Option error: %s" % str(e))
        Alert.send_alert(Alert.MsgCode.FAILOVER_ERROR, fsname=fs_global, detail=str(e))
        sys.exit(1)
    except KeyboardInterrupt:
        print "Interrupted."
        log_warning("Interrupted.")
        Alert.send_alert(Alert.MsgCode.FAILOVER_ERROR, fsname=fs_global, detail="Interrupted by a signal.")
        sys.exit(2)
    except Config.BaseError, e:
        print "Unable to load script configuration: %s" % str(e)
        Alert.send_alert(Alert.MsgCode.FAILOVER_ERROR, fsname=fs_global, detail="Unable to load script configuration file.")
        sys.exit(10)
    except Exceptions.ConfigurationError, e:
        log_error("Unexpected configuration: %s" % str(e))
        Alert.send_alert(Alert.MsgCode.FAILOVER_ERROR, fsname=fs_global, detail=str(e))
        sys.exit(20)
    except Exceptions.ConnectionError, e:
        log_error("A connection error occurred: %s" % str(e))
        Alert.send_alert(Alert.MsgCode.FAILOVER_ERROR, fsname=fs_global, detail=str(e))
        sys.exit(70)
    except Exceptions.OperationError, e:
        log_error("An operation error occurred: %s" % str(e))
        Alert.send_alert(Alert.MsgCode.FAILOVER_ERROR, fsname=fs_global, detail=str(e))
        sys.exit(90)
    except EnvironmentError, e:
        log_exception("A system error occurred: %s", e)
        Alert.send_alert(Alert.MsgCode.FAILOVER_ERROR, fsname=fs_global, detail=str(e))
        sys.exit(99)
    except Exception, e:
        log_exception("An internal error occurred: %s", e)
        Alert.send_alert(Alert.MsgCode.FAILOVER_ERROR, fsname=fs_global, detail=str(e))
        sys.exit(99)

    # Do not send an "OK" alert here because this is just job registration.
    sys.exit(0)
