#!/usr/bin/python

BASEDIR = "/home/nasroot/foscripts"
PROGRAM = "setup_pair"

import sys
import os

sys.path.append(os.path.join(BASEDIR, "bin/folib"))
import FailoverConfig
from CommonLogger import log_debug, log_info, log_warning, log_error, log_exception, setup_rotation_logger
import Utility

sys.path.append(os.path.join(BASEDIR, "bin"))
import Config
import Exceptions
import getinfo_common
import setup_common
import setup_pair_group

def input_pair_nodes():
    print
    print ">>> Specify an IP address of primary/secondary node."
    print

    primary = Utility.my_input("Primary node IP address: ")
    if not Utility.is_IP_address(primary):
        log_warning("Input data is not an IP address: %s" % primary)
        print "Input data is not an IP address."
        raise Exceptions.NoExecution()
    secondary = Utility.my_input("Secondary node IP address: ")
    if not Utility.is_IP_address(secondary):
        log_warning("Input data is not an IP address: %s" % secondary)
        print "Input data is not an IP address."
        raise Exceptions.NoExecution()

    log_info("Input pair: primary=%s, secondary=%s" % (primary, secondary))
    return (primary, secondary)

def select_file_systems_for_failover(primary, config, ssh_config):
    """ Ask a user to enter file system(s) to failover.
    Returns [] if 'All' is selected.
    Otherwise returns a list of file system names that are selected.
    """

    ALL = 1
    SELECTED = 2

    log_info("Selecting a file system(s) to be failed over ...")

    print
    print ">>> Specify a file system(s) on the primary node to be failed over."
    print

    menu = [(ALL, "All file systems"), (SELECTED, "Only selected file system(s)")]
    selected_list = Utility.select_items("Select", menu, onlyone=True)
    if selected_list is None:
        log_warning("Selecting a file system did not succeed.")
        print "Process canceled."
        raise Exceptions.NoExecution()
    selected = selected_list[0]
    if selected == ALL:
        log_info("All is selected.")
        result = []
    elif selected == SELECTED:
        # Ask the user to select a file system(s) to be failed over

        fs_list = getinfo_common.get_remote_file_systems(primary, config, ssh_config)
        if len(fs_list) == 0:
            log_warning("No file systems found on the primary node (%s)." % primary)
            print "No file systems found on the primary node (%s). Proceed with an assumption that all file systems will be failed over." % primary
            result = []
        else:
            fsname_list = [fs.fileSystemName for fs in fs_list]
            log_info("File systems on %s: %s" % (primary, fsname_list))
            fs_menu = []
            for idx, fsname in enumerate(fsname_list):
                fs_menu.append((idx + 1, fsname))
            selected_list = Utility.select_items("Select file system(s) to be failed over (separate with comma for multiple selection)", fs_menu, title="File systems on %s" % primary)
            if selected_list is None:
                log_error("Selecting a file system(s) did not succeed.")
                print "Process canceled."
                raise Exceptions.NoExecution()

            result = []
            for selected in selected_list:
                result.append(fsname_list[selected - 1])

    log_info("Selected: %s" % result)
    return result

def add_pair(pairconf, stconf, config, ssh_config):
    log_info("Start adding a pair.")

    # Enter IP addresses of nodes of the pair
    primary, secondary = input_pair_nodes()

    # Check if primary == secondary
    if primary == secondary:
        print "The same IP address is specified for both primary and secondary."
        log_warning("The same IP address is specified for both primary and secondary. primary=%s, secondary=%s" % (primary, secondary))
        raise Exceptions.NoExecution()

    # Check if the pair is already registered
    if pairconf.exists(primary=primary, secondary=secondary):
        print "The entered pair is already registered."
        log_warning("The entered pair is already registered. primary=%s, secondary=%s" % (primary, secondary))
        raise Exceptions.NoExecution()

    # Check if the primary node is already registered as a secondary node (and vice versa)
    if pairconf.exists(secondary=primary):
        print "The entered primary node is already registered as a secondary node."
        log_warning("The entered primary node is already registered as a secondary node. Entered primary=%s, secondary=%s" % (primary, secondary))
        raise Exceptions.NoExecution()
    if pairconf.exists(primary=secondary):
        print "The entered secondary node is already registered as a primary node."
        log_warning("The entered secondary node is already registered as a primary node. Entered primary=%s, secondary=%s" % (primary, secondary))
        raise Exceptions.NoExecution()

    # Select a file system(s) to be failed over
    fsname_list = select_file_systems_for_failover(primary, config, ssh_config)

    # Add the pair
    is_all = True if len(fsname_list) == 0 else False
    target_fslist = None if is_all else fsname_list
    print
    print "Adding the pair (primary: %s, secondary: %s) ..." % (primary, secondary)
    pid = pairconf.add(primary, secondary, target_fslist=target_fslist, exclude_fslist=None, is_all=is_all)

    # Setup default storage policy (Auto - usage with 100 GB margin)
    default_policy = FailoverConfig.StoragePolicy(FailoverConfig.StoragePolicy.FSDEF, policy_string=config.default_policy_string)
    print "Adding a default storage policy (%s) for the new pair ..." % default_policy.to_display_string()
    log_info("Adding a default storage policy (%s) for the new pair ..." % default_policy.to_display_string())
    stconf.add(pid, policy=default_policy)

    print
    print "Finished."

def edit_pair(pairconf, stconf, config, ssh_config):
    """ Edit a pair.
    Only target file systems can be changed.
    Delete and Add is required to change the pair itself.
    """
    log_info("Start editing a pair.")

    pid = setup_common.select_pair(pairconf)
    if pid is None:
        print
        print "No pairs are selected."
        log_warning("No pairs are selected.")
        return

    pair_list = pairconf.get_list(pid)
    pair = pair_list[0][1]
    log_info("Selected pair: %s" % pair)

    # Display the current defined file systems to be failed over
    print
    print "File systems to be failed over on the pair:"
    if pair.is_all:
        print "\t(All)"
    else:
        for fs in pair.target_fslist:
            print "\t%s" % fs

    # Select change it or not
    answer = Utility.yes_or_no("Do you want to change file systems to be failed over?")
    if answer is False:
        print "Process canceled."
        log_info("Process canceled.")
        return

    # Change the file systems
    fsname_list = select_file_systems_for_failover(pair.primary, config, ssh_config)

    print "Updating the pair ..."
    is_all = True if len(fsname_list) == 0 else False
    target_fslist = None if is_all else fsname_list
    pairconf.edit(pid, target_fslist=target_fslist, exclude_fslist=None, is_all=is_all)

    # Remove storage policy for the file systems that will not be failed over
    # Default storage policy will be applied to the newly selected file systems
    print "Updating storage policies for the updated pair ..."
    log_info("Updating storage policies for the updated pair ...")
    if not is_all:
        stconf_list = stconf.get_list(pid)
        for fsname in [el[1] for el in stconf_list]:
            if fsname != FailoverConfig.StoragePolicy.FSDEF and fsname not in fsname_list:
                stconf.delete(pid, fsname)
        # Need to change default policy to the global default policy if it's Like for Like
        # because Like for Like can be applied only when "(All)" is selected.
        default_policy = [el[2] for el in stconf_list if el[1] == FailoverConfig.StoragePolicy.FSDEF][0]
        if default_policy.policy == FailoverConfig.StoragePolicy.LIKEFORLIKE:
            print "--------"
            print "NOTE:"
            print "Default policy is changed to global default policy because 'Like for Like' cannot be applied when a specific file system(s) is selected."
            print "--------"
            new_default_policy = FailoverConfig.StoragePolicy(FailoverConfig.StoragePolicy.FSDEF, policy_string=config.default_policy_string)
            stconf.delete(pid, FailoverConfig.StoragePolicy.FSDEF)
            stconf.add(pid, policy=new_default_policy)

    print
    print "Finished."

def delete_pair(pairconf, stconf, pgconf):
    """ Delete a pair.
    """
    log_info("Start deleting a pair ...")

    pid = setup_common.select_pair(pairconf)
    if pid is None:
        print
        print "No pairs are selected."
        log_warning("No pairs are selected.")
        return

    pair_list = pairconf.get_list(pid)
    pair = pair_list[0][1]
    log_info("Selected pair: %s" % pair)
    answer = Utility.yes_or_no("Do you want to delete the pair (primary: %s, secondary: %s)?" % (pair.primary, pair.secondary))
    if answer is False:
        print "Process canceled."
        log_info("Process canceled.")
        return

    # Delete the selected pair
    print "Deleting the pair ..."
    pairconf.delete(pair_id=pid)

    # Delete the storage policies for the deleted pair
    print "Deleting storage policies for the deleted pair ..."
    stconf.delete(pid)

    # Delete the pair from pair groups
    print "Deleting the pair from pair groups ..."
    setup_pair_group.delete_pair_from_all_pair_groups(pid, pgconf)

    print
    print "Finished."

def save_configuration(pairconf, stconf, pgconf):
    # Save the configuration
    print
    print "Saving the configuration ..."
    pairconf.save()
    stconf.save()
    pgconf.save()

def main():
    # Enable Ctrl-C interruption
    Utility.enable_interrupt()

    # Load configs
    config = Config.Config()

    # Setup logging
    log_file_path = os.path.join(config.log_dir, config.common_log_file)
    setup_rotation_logger(__name__, config.log_level, log_file_path, config.log_size, config.log_gen)

    log_info("----------------------- %s --------------------------" % PROGRAM)

    # Load the existing pair configuration
    log_info("Getting list of pair configuration.")
    try:
        confgroup_list = FailoverConfig.list_configuration_group(config.fo_conf_dir)
    except FailoverConfig.ConfigurationError, e:
        log_error("Failed to count number of configurations due to an configuration error. error=%s" % str(e))
        raise Exceptions.ConfigurationError(str(e))

    # Select a configuration group
    confgroup = setup_common.select_configuration_group(confgroup_list)

    # Load pair configuration and storage policy configuration
    pairconf = FailoverConfig.PairConfig(config.fo_conf_dir, confgroup)
    stconf = FailoverConfig.StoragePolicyConfig(config.fo_conf_dir, confgroup)
    pgconf = FailoverConfig.PairGroupConfig(config.fo_conf_dir, confgroup)

    # Setup SSHConfig
    ssh_config = Config.SSHConfig(config.sys_conf_dir, pairconf, confgroup)

    # Main menu
    OP_ADD = 1
    OP_EDIT = 2
    OP_DELETE = 3
    OP_DISPLAY = 4
    OP_EXIT = 0
    main_menu = [
        (OP_ADD, "Add a pair"),
        (OP_EDIT, "Edit a pair"),
        (OP_DELETE, "Delete a pair"),
        (OP_DISPLAY, "Display pairs"),
        (OP_EXIT, "Exit")
    ]

    while True:
        # Select an operation
        selected = setup_common.select_operation(main_menu)

        try:
            # Execute the selected operation
            if selected == OP_DISPLAY:
                print
                print "----- Display pairs -----"
                setup_common.display_pairs(pairconf, stconf)
            elif selected == OP_ADD:
                print
                print "----- Add a pair -----"
                add_pair(pairconf, stconf, config, ssh_config)
                save_configuration(pairconf, stconf, pgconf)
            elif selected == OP_EDIT:
                print
                print "----- Edit a pair -----"
                edit_pair(pairconf, stconf, config, ssh_config)
                save_configuration(pairconf, stconf, pgconf)
            elif selected == OP_DELETE:
                print
                print "----- Delete a pair -----"
                delete_pair(pairconf, stconf, pgconf)
                save_configuration(pairconf, stconf, pgconf)
            elif selected == OP_EXIT:
                log_info("Exit selected.")
                break
        except Exceptions.NoExecution:
            print "Going back to the main menu ..."

        print
        notused = raw_input(">>> Press Enter to go back to the main menu.")

if __name__ == '__main__':
    try:
        main()
    except Exceptions.NoExecution:
        sys.exit(0)
    except KeyboardInterrupt:
        print "Interrupted."
        log_warning("Interrupted.")
        sys.exit(2)
    except Config.BaseError, e:
        print "Unable to load script configuration: %s" % str(e)
        sys.exit(10)
    except Exceptions.ConfigurationError, e:
        log_error("Unexpected configuration found: %s" % str(e))
        sys.exit(20)
    except Exceptions.ConnectionError, e:
        log_error("A connection error occurred: %s" % str(e))
        sys.exit(70)
    except Exceptions.OperationError, e:
        log_error("An operation error occurred: %s" % str(e))
        sys.exit(90)
    except EnvironmentError, e:
        log_exception("A system error occurred: %s", e)
        sys.exit(99)
    except Exception, e:
        log_exception("An internal error occurred: %s", e)
        sys.exit(99)

    sys.exit(0)
