#!/usr/bin/python

BASEDIR = "/home/nasroot/foscripts"
PROGRAM = "setup_pair_group"

import sys
import os
import re

sys.path.append(os.path.join(BASEDIR, "bin/folib"))
import FailoverConfig
from CommonLogger import log_debug, log_info, log_warning, log_error, log_exception, setup_rotation_logger
import Utility

sys.path.append(os.path.join(BASEDIR, "bin"))
import Config
import Exceptions
import setup_common

def display_pair_groups(pgconf, pairconf):
    """ Display registered pair groups.

    This function displays all registered pair group names and pairs for each pair group.

    Args:
        pgconf (FailoverConfig.PairGroupConfig): PairGroup configuration.
        pairconf (FailoverConfig.PairConfig): Pair configuration.
    Returns:
        None
    """
    pair_group_list = pgconf.get_list()
    if len(pair_group_list) == 0:
        print
        print "*** There is no pair groups defined. ***"
    else:
        for name, pair_group in pair_group_list:
            print
            print "%s:" % name
            for pid in pair_group.pair_list:
                if not pairconf.exists(pair_id=pid):
                    log_warning("A pair registered in pair group %s does not exist in pair configuration. (Pair ID: %s)" % (name, pid))
                    continue
                pairs = pairconf.get_list(pair_id=pid)
                notused, pair = pairs[0]
                print "\t%s - %s" % (pair.primary, pair.secondary)
    print

def add_pair_group(pgconf, pairconf):
    """ Add a new pair group.

    Args:
        pgconf (FailoverConfig.PairGroupConfig): Pair group configuration.
        pairconf (FailoverConfig.PairConfig): Pair configuraiton
    Returns:
        None. Directly update pgconf.
    """

    log_info("Start adding a pair group.")

    # Enter name of a new pair group
    name = input_pair_group_name()

    # Check if the name is already registered
    if pgconf.exists(name):
        print "The name is already used. Specify a different name."
        log_info("The entered name is already used. name: %s" % name)
        raise Exceptions.NoExecution()

    # Select a pair(s) to add to the group
    pid_list = setup_common.select_pair(pairconf, msg="Select a pair(s) to add to the group (separate with comma for multiple selection)", onlyone=False)
    if pid_list is None or len(pid_list) == 0:
        print
        print "No pairs are selected."
        log_warning("No pairs are selected.")
        raise Exceptions.NoExecution()

    print
    print "Adding \"%s\" as a new pair group ..." % name
    pgconf.add(name, pid_list)

    print
    print "Finished."

def input_pair_group_name():
    print
    print ">>> Enter a new pair group name."
    print

    name = Utility.my_input("Pair group name: ")
    if name is None:
        log_info("Nothing is input for a new pair group name.")
        return None

    if not is_valid_pair_group_name(name):
        print "The name is not valid as a pair group name. name: %s" % name
        print "Characters that can be used fo pair group name: number/alphabet/underscore/white space/dash"
        log_info("The name is not valid as a pair group name. name: %s" % name)
        raise Exceptions.NoExecution()

    log_info("Input pair group name: %s" % name)
    return name

def is_valid_pair_group_name(name):
    """ Check if the specified name is valid as a pair group name.

    These characters can be included in a pair group name.
        Number
        Alphabet
        Underscore
        White space
        Dash

    Args:
        name (str): Name to be checked.
    Returns:
        bool: True if it's valid name. False otherwise.
    """
    m = re.match(r"^[0-9a-zA-Z_ -]+$", name)
    if m is None:
        return False
    else:
        return True

def edit_pair_group(pgconf, pairconf):
    """ Edit a pair group.

    This function does either of the two operations below.
        Add a pair(s) to a pair group.
        Delete a pair(s) from a pair group.

    Args:
        pgconf (FailoverConfig.PairGroupConfig): Pair group configuraiton.
        pairconf (FailoverConfig.PairConfig): Pair configuration.
    Returns:
        None. Directly updates pgconf.
    """
    log_info("Start editing a pair group.")

    # Select a pair group
    print
    print ">>> Select a pair group to edit."
    name = select_pair_group(pgconf)

    # Select an operation for the pair group (Add or Delete)
    OP_ADD = 1
    OP_DELETE = 2
    OP_DISPLAY = 3
    OP_EXIT = 0

    op_menu = [
        (OP_ADD, "Add a pair to the pair group"),
        (OP_DELETE, "Delete a pair from the pair group"),
        (OP_DISPLAY, "Display pairs that belong to the pair group"),
        (OP_EXIT, "Finish")
    ]
    while True:
        selected = setup_common.select_operation(op_menu)
        try:
            if selected == OP_ADD:
                add_pair_to_pair_group(name, pgconf, pairconf)
            elif selected == OP_DELETE:
                delete_pair_from_pair_group(name, pgconf, pairconf)
                # Check if the pair group was deleted by being empty
                if not pgconf.exists(name):
                    # Go back to the main menu because this pair group no longer exists.
                    print
                    print "Going back to the main menu ..."
                    break
            elif selected == OP_DISPLAY:
                display_pairs_in_pair_group(name, pgconf, pairconf)
            elif selected == OP_EXIT:
                log_info("Exit from edit menu.")
                break
        except Exceptions.NoExecution:
            print "Going back to edit menu ..."

        print
        notused = raw_input(">>> Press Enter to go back to edit menu.")

def add_pair_to_pair_group(name, pgconf, pairconf):
    log_info("Start adding a pair(s) to pair group \"%s\"." % name)

    # Select a pair(s) to add to the group
    print
    print ">>> Select a pair(s) to add to the pair group."
    pid_list = setup_common.select_pair(pairconf, msg="Select a pair(s) to add to the group (separate with comma for multiple selection)", onlyone=False)
    if pid_list is None or len(pid_list) == 0:
        print
        print "No pairs are selected."
        log_warning("No pairs are selected.")
        raise Exceptions.NoExecution()
    log_info("Selected pairs: %s" % pid_list)

    # Add the selected pair to the group
    print
    print "Adding the pair(s) to pair group \"%s\" ..." % name
    pg_list = pgconf.get_list(name=name)
    notused, pairgroup = pg_list[0]
    new_pid_list = list(set(pid_list + pairgroup.pair_list))
    pgconf.update(name, new_pid_list)

    print
    print "Finished."

def delete_pair_from_pair_group(name, pgconf, pairconf):
    log_info("Start deleting a pair(s) from pair group \"%s\"." % name)

    # Select a pair(s) to delete from the group.
    print
    print ">>> Select a pair(s) to delete from the pair group."
    pid_list_to_be_deleted = select_pair_in_pair_group(name, pgconf, pairconf)

    # Delete the selected pair(s) from the group.
    _delete_pair_from_pair_group(pid_list_to_be_deleted, name, pgconf)

    print
    print "Finished."

def _delete_pair_from_pair_group(pid_list_to_be_deleted, name, pgconf):
    print
    print "Deleting the pair(s) from pair group \"%s\" ..." % name
    pg_list = pgconf.get_list(name=name)
    notused, pairgroup = pg_list[0]
    current_pid_list = pairgroup.pair_list
    new_pid_list = list(set(current_pid_list) - set(pid_list_to_be_deleted))
    if len(new_pid_list) == 0:
        print "Deleting the pair group \"%s\" because there is no pairs in the group ..." % name
        pgconf.delete(name)
    else:
        pgconf.update(name, new_pid_list)

def pair_exists_in_pair_group(pair_id, pairgroup):
    return pair_id in pairgroup.pair_list

def delete_pair_from_all_pair_groups(pair_id, pgconf):
    for name, pairgroup in pgconf.get_list():
        if pair_exists_in_pair_group(pair_id, pairgroup):
            _delete_pair_from_pair_group([pair_id], name, pgconf)

def select_pair_in_pair_group(name, pgconf, pairconf):
    """ Select a pair(s) from the list of pairs in a pair group.

    This function displays a list of pairs in the specified pair group,
    and let user select a pair(s) from the list.
    Returns a list of pair IDs that are selected.

    Args:
        name (str): Name of the pair group.
        pgconf (FailoverConfig.PairGroupConfig): Pair group configuration.
        pairconf (FailoverConfig.PairConfig): Pair configuration.
    Returns:
        List of int: List of pair ID that are selected.
    """

    # Prepare menu
    pg_list = pgconf.get_list(name=name)
    name, pairgroup = pg_list[0]
    menu = []
    idx_to_pid = {}
    for i, pid in enumerate(pairgroup.pair_list):
        idx = i + 1
        pair_list = pairconf.get_list(pair_id=pid)
        if len(pair_list) == 0:
            log_info("A pair is not found in pair configuration though it is listed in a pair group. pair ID: %s, pair group name: %s" % (pid, name))
            continue
        notused, pair = pair_list[0]
        menu.append((idx, "%s - %s" % (pair.primary, pair.secondary)))
        idx_to_pid[idx] = pid

    # Select a pair(s)
    selected_list = Utility.select_items("Select a pair(s) to add to the group (separate with comma for multiple selection)", menu, title="Pairs belong to pair group \"%s\"" % name)
    if selected_list is None:
        log_warning("Selecting a pair(s) did not succeed.")
        print "Process canceled."
        raise Exceptions.NoExecution()

    # Convert menu index to pair ID
    pid_list = [idx_to_pid[idx] for idx in selected_list]

    return pid_list

def display_pairs_in_pair_group(name, pgconf, pairconf):
    """ Display pairs that belong to a pair group.

    This function displays pairs that belong to a pair group.

    Args:
        name (str): Name of a pair group that its pairs are displayed.
        pgconf (FailoverConfig.PairGroupConfig): Pair group configuration.
        pairconf (FailoverConfig.PairConfig): Pair configuration.
    Returns:
        None
    """
    pg_list = pgconf.get_list(name=name)
    name, pairgroup = pg_list[0]
    print
    print "Pairs belong to pair group \"%s\":" % name
    for pid in pairgroup.pair_list:
        pair_list = pairconf.get_list(pair_id=pid)
        if len(pair_list) == 0:
            log_info("A pair is not found in pair configuration though it is listed in a pair group. pair ID: %s, pair group name: %s" % (pid, name))
            continue
        notused, pair = pair_list[0]
        print "\t%s - %s" % (pair.primary, pair.secondary)

def select_pair_group(pgconf):
    """ Select a pair group from the defined groups.

    This function displays defined pair groups and let user select a pair
    group.
    Returns the name of the selected pair group.

    Args:
        pgconf (FailoverConfig.PairGroupConfig): Pair group configruation.
    Returns:
        str: Name of the selected pair group.
    """

    # Build a menu to select a pair group
    menu = []
    pg_list = pgconf.get_list()
    for i, (pgname, notused) in enumerate(pg_list):
        idx = i + 1
        menu.append((idx, pgname))

    selected_list = Utility.select_items("Select a pair group", menu, title="Defined pair groups", onlyone=True)
    if selected_list is None:
        log_warning("Selecting a pair group did not succeed.")
        print "Process canceled."
        raise Exceptions.NoExecution()
    selected = selected_list[0]
    name, notused = pg_list[selected - 1]
    log_info("Selected pair group: %s" % name)

    return name

def delete_pair_group(pgconf):
    log_info("Start deleting a pair group ...")

    # Select a pair group
    print
    print ">>> Select a pair group to delete."
    name = select_pair_group(pgconf)

    # Confirmation
    print
    answer = Utility.yes_or_no("Do you want to delete the pair group \"%s\"?" % name)
    if answer is False:
        print "Process canceled."
        log_info("Process canceled.")
        raise Exceptions.NoExecution()

    # Delete the selected pair group
    print
    print "Deleting pair group \"%s\" ..." % name
    pgconf.delete(name)

    print
    print "Finished."

def save_configuration(pgconf):
    # Save the configuration
    print
    print "Saving the configuration ..."
    pgconf.save()

def main():
    # Enable Ctrl-C interruption
    Utility.enable_interrupt()

    # Load configs
    config = Config.Config()

    # Setup logging
    log_file_path = os.path.join(config.log_dir, config.common_log_file)
    setup_rotation_logger(__name__, config.log_level, log_file_path, config.log_size, config.log_gen)

    log_info("----------------------- %s --------------------------" % PROGRAM)

    # Load the existing pair configuration
    log_info("Getting list of pair configuration.")
    try:
        confgroup_list = FailoverConfig.list_configuration_group(config.fo_conf_dir)
    except FailoverConfig.ConfigurationError, e:
        log_error("Failed to count number of configurations due to an configuration error. error=%s" % str(e))
        raise Exceptions.ConfigurationError(str(e))

    # Select a configuration group
    confgroup = setup_common.select_configuration_group(confgroup_list)

    # Load pair configuration and pair group configuration
    pairconf = FailoverConfig.PairConfig(config.fo_conf_dir, confgroup)
    pgconf = FailoverConfig.PairGroupConfig(config.fo_conf_dir, confgroup)
    config.pairconf = pairconf

    # Main menu
    OP_ADD = 1
    OP_EDIT = 2
    OP_DELETE = 3
    OP_DISPLAY = 4
    OP_EXIT = 0
    main_menu = [
        (OP_ADD, "Add a pair group"),
        (OP_EDIT, "Edit a pair group"),
        (OP_DELETE, "Delete a pair group"),
        (OP_DISPLAY, "Display pair groups"),
        (OP_EXIT, "Exit")
    ]

    while True:
        # Select an operation
        selected = setup_common.select_operation(main_menu)

        try:
            # Execute the selected operation
            if selected == OP_DISPLAY:
                print
                print "----- Display pair groups -----"
                display_pair_groups(pgconf, pairconf)
            elif selected == OP_ADD:
                print
                print "----- Add a pair group -----"
                add_pair_group(pgconf, pairconf)
                save_configuration(pgconf)
            elif selected == OP_EDIT:
                print
                print "----- Edit a pair group -----"
                edit_pair_group(pgconf, pairconf)
                save_configuration(pgconf)
            elif selected == OP_DELETE:
                print
                print "----- Delete a pair group -----"
                delete_pair_group(pgconf)
                save_configuration(pgconf)
            elif selected == OP_EXIT:
                log_info("Exit selected.")
                break
        except Exceptions.NoExecution:
            print "Going back to the main menu ..."

        print
        notused = raw_input(">>> Press Enter to go back to the main menu.")

if __name__ == '__main__':
    try:
        main()
    except Exceptions.NoExecution:
        sys.exit(0)
    except KeyboardInterrupt:
        print "Interrupted."
        log_warning("Interrupted.")
        sys.exit(2)
    except Config.BaseError, e:
        print "Unable to load script configuration: %s" % str(e)
        sys.exit(10)
    except Exceptions.ConfigurationError, e:
        log_error("Unexpected configuration: %s" % str(e))
        sys.exit(20)
    except EnvironmentError, e:
        log_exception("A system error occurred: %s", e)
        sys.exit(99)
    except Exception, e:
        log_exception("An internal error occurred: %s", e)
        sys.exit(99)

    sys.exit(0)
